#!/usr/bin/perl -w

# Copyright 2016, 2017, 2018, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.

use strict;
use 5.006;
use warnings;
use Carp 'croak';
use FindBin;
use List::Util 'sum';
use Math::BaseCnv 'cnv';
use Math::PlanePath::TerdragonCurve;
use Math::NumSeq::PlanePathTurn;
use FLAT;
use Test::More;

use lib $FindBin::Bin;
use MyTerdragon ':all';
use MyFLAT ':all';

# uncomment this to run the ### lines
# use Smart::Comments;

plan tests => 277;


my $all_ternary = FLAT::Regex->new('(0|1|2)*')->as_dfa
  ->MyFLAT::set_name("all ternary");
my $zeros_flat = FLAT::Regex->new('0*')->as_dfa
  ->MyFLAT::set_name("zeros");

#------------------------------------------------------------------------------
# Turn Tree

# turn tree is repeated 4-star replacement
require Graph;
require MyGraphs;
foreach my $k (1 .. 6) {
  my $rep = Graph->new (undirected => 1);
  $rep->add_vertex(0);
  foreach (2 .. $k) {
    $rep = MyGraphs::Graph_star_replacement($rep, 4, edges_between=>0);
  }
  my $turn = triangle_turn_tree_Graph($k);
  my $ok = MyGraphs::Graph_is_isomorphic($rep, $turn);
  ok ($ok, "triangle_turn_tree_Graph() vs Graph_star_replacement(), k=$k");
  if (! $ok) {
    # MyGraphs::Graph_view($turn);
    # MyGraphs::Graph_view($rep);
  }
}

#------------------------------------------------------------------------------
# dNextLsub1_flat

foreach my $d (1 .. 3) {
  my $dNextL = dNextL_flat($d);
  my $dNextLsub1 = dNextLsub1_flat($d);
  my $by_incr = $dNextL->MyFLAT::digits_increment(add => -1)
    ->MyFLAT::optional_leading_0s
    ->MyFLAT::set_name("dNextL incr -1");
  ok ($dNextLsub1->equals($by_incr));
  # MyFLAT::FLAT_check_is_equal($dNextLsub1,$by_incr);
}


#------------------------------------------------------------------------------
# dNextL, dNextR

# high 0s are optional in dNextL_flat and dNextR_flat, so ensure optional in
# the Regex
#
ok(dNextL_flat(1)->equals(FLAT::Regex->new('((0|1|2)* 0 | []) 2*')->as_dfa));
ok(dNextL_flat(2)->equals(FLAT::Regex->new('((0|1|2)* 0 | []) 2* 1 
                                          |  (0|1|2)* 1       2* 2')->as_dfa));
ok(dNextL_flat(3)->equals(FLAT::Regex->new('(0|1|2)* 1 2* 1')->as_dfa));

ok (dNextR_flat(1)->equals(FLAT::Regex->new('(0|1|2)* 1 2*')->as_dfa));
ok (dNextR_flat(2)->equals(FLAT::Regex->new('(0|1|2)* 0 | []')->as_dfa));
ok (dNextR_flat(3)->equals(FLAT::Regex->new('((0|1|2)* 0 | []) 2* 2')->as_dfa));

foreach my $func (\&dNextL_flat, \&dNextR_flat) {
  ok ($func->(1)->intersect($func->(2))->is_empty);
  ok ($func->(1)->intersect($func->(3))->is_empty);
  ok ($func->(2)->intersect($func->(3))->is_empty);

  my $u = $func->(1) ->union($func->(2)) ->union($func->(3))->as_dfa
    ->MyFLAT::set_name("union");
  ok ($u->equals($all_ternary));
  # MyFLAT::FLAT_check_is_equal($u,$all);
}
{
  my $f1 = TurnLpred_flat()->MyFLAT::digits_increment(add => -1)
    ->MyFLAT::optional_leading_0s
    ->MyFLAT::set_name('dNextL=1 by incr');
  ok ($f1->equals(dNextL_flat(1)));

  my $f_next_R = TurnRpred_flat()->MyFLAT::digits_increment(add => -1)
    ->MyFLAT::optional_leading_0s
    ->MyFLAT::set_name('TurnRpredNext');

  my $f2 = $f1->MyFLAT::digits_increment(add => -1)
    ->intersect($f_next_R)
    ->MyFLAT::optional_leading_0s
    ->MyFLAT::set_name('dNextL=2 by incr');
  ok ($f2->equals(dNextL_flat(2)));

  my $f3 = $f2->MyFLAT::digits_increment(add => -1)
    ->intersect($f_next_R)
    ->MyFLAT::optional_leading_0s
    ->MyFLAT::set_name('dNextL=3 by incr');
  ok ($f3->equals(dNextL_flat(3)));
}

{
  my $bad = 0;
  my @f = map {dNextL_flat($_)} 1..3;
  foreach my $n (0 .. 500) {
    my $by_search = dNextL_by_search($n);
    my $by_func   = dNextL($n);
    my $n3 = cnv($n,10,3);
    my @by_flat = map {$_->contains($n3)} @f;
    unless (sum(@by_flat) == 1) {
      last if $bad++ > 10;
    }
    my $by_flat = $by_flat[0] ? 1 : $by_flat[1] ? 2 : 3;
    unless ($by_func == $by_search) {
      diag "n=$n dNextL()=$by_func search=$by_search";
      last if $bad++ > 10;
    }
    unless ($by_flat == $by_search) {
      last if $bad++ > 10;
    }
  }
  is($bad, 0);
}
sub dNextL_by_search {
  my ($n) = @_;
  foreach my $d (1 .. 3) {
    if (TurnLpred($n+$d)) { return $d; }
  }
  die "oops";
}

#------
ok (dNextR_flat(1)->contains('1'));
ok (dNextR_flat(3)->contains('2'));

{
  my $bad = 0;
  my @f = map {dNextR_flat($_)} 1..3;
  foreach my $n (0 .. 500) {
    my $by_search = dNextR_by_search($n);
    my $by_func   = dNextR($n);
    my $n3 = cnv($n,10,3);
    my @by_flat = map {$_->contains($n3)} @f;
    unless (sum(@by_flat) == 1) {
      last if $bad++ > 10;
    }
    my $by_flat = $by_flat[0] ? 1 : $by_flat[1] ? 2 : 3;
    unless ($by_func == $by_search) {
      diag "n=$n dNextR()=$by_func search=$by_search";
      last if $bad++ > 10;
    }
    unless ($by_flat == $by_search) {
      diag "n=$n dNextR_flat()=$by_flat search=$by_search";
      last if $bad++ > 10;
    }
  }
  is($bad, 0);
}
sub dNextR_by_search {
  my ($n) = @_;
  foreach my $d (1 .. 3) {
    if (TurnRpred($n+$d)) { return $d; }
  }
  die "oops";
}


#------------------------------------------------------------------------------
# TurnLpred(), TurnRpred()

ok(! TurnLpred_flat() ->contains(''));
ok(! TurnRpred_flat()->contains(''));
ok(TurnLpred_flat() ->intersect(zero_digits_flat(3))->is_empty,
   "TurnLpred_flat() no match zero");
ok(TurnRpred_flat()->intersect(zero_digits_flat(3))->is_empty,
   "TurnRpred_flat() no match zero");
ok(TurnLpred_flat() ->intersect(TurnRpred_flat())->is_empty,
   "TurnLpred_flat() and TurnRpred_flat() disjoint");

is (TurnLpred(0), 0);
is (TurnRpred(0), 0);
{
  my $path = Math::PlanePath::TerdragonCurve->new;
  my $seq = Math::NumSeq::PlanePathTurn->new(planepath_object=>$path,
                                             turn_type => 'Left');
  my $bad = 0;
  foreach (1 .. 500) {
    my ($n,$by_seq) = $seq->next;
    my $by_func = TurnLpred($n);
    my $n3 = cnv($n,10,3);
    my $by_flat = TurnLpred_flat->contains($n3);
    unless ($by_seq == $by_func) {
      diag "n=$n TurnLpred()=$by_func seq=$by_seq";
      last if $bad++ > 10;
    }
    unless ($by_seq == $by_flat) {
      diag "n=$n TurnLpred_flat()=$by_flat seq=$by_seq";
      last if $bad++ > 10;
    }
  }
  is($bad, 0);
}
{
  my $path = Math::PlanePath::TerdragonCurve->new;
  my $seq = Math::NumSeq::PlanePathTurn->new(planepath_object=>$path,
                                             turn_type => 'Right');
  my $bad = 0;
  foreach (1 .. 500) {
    my ($n,$by_seq) = $seq->next;
    my $by_func = TurnRpred($n);
    my $n3 = cnv($n,10,3);
    my $by_flat = TurnRpred_flat->contains($n3);
    unless ($by_seq == $by_func) {
      diag "n=$n TurnRpred()=$by_func seq=$by_seq";
      last if $bad++ > 10;
    }
    unless ($by_seq == $by_flat) {
      diag "n=$n TurnRpred_flat()=$by_flat seq=$by_seq";
      last if $bad++ > 10;
    }
  }
  is($bad, 0);
}


#------------------------------------------------------------------------------
# MyTerdragonConf expand()

{
  my $got = MyTerdragonConf->initial->expand(0);
  my $want = MyTerdragonConf->new
    (segs => [ '0,0,0', '2,0,1', '1,1,0' ]);
  ok($got->equals($want));
}
{
  my $got = MyTerdragonConf->initial->expand(1);
  my $want = MyTerdragonConf->new
    (segs => [ '1,1,2', '0,0,0', '2,0,2' ]);
  ok($got->equals($want));
}
{
  my $got = MyTerdragonConf->initial->expand(2);
  my $want = MyTerdragonConf->new
    (segs => [ '-1,-1,0', '1,-1,1', '0,0,0' ]);
  ok($got->equals($want));
}

{
  my $got = MyTerdragonConf->new(segs=>['0,0,0'])->expand(0, 1);  # alt
  my $want = MyTerdragonConf->new
    (segs => [ '0,0,0', '2,0,2', '1,-1,0' ]);
  ok ($got->equals($want));
}
{
  my $got = MyTerdragonConf->new(segs=>['0,0,0'])->expand(1, 1);  # alt
  my $want = MyTerdragonConf->new
    (segs => [ '1,-1,1', '0,0,0', '2,0,1' ]);
  ok ($got->equals($want));
}
{
  my $got = MyTerdragonConf->new(segs=>['0,0,0'])->expand(2, 1);  # alt
  my $want = MyTerdragonConf->new
    (segs => [ '-1,1,0', '1,1,2', '0,0,0' ]);
  ok ($got->equals($want));
}


# {
#   my $conf = MyTerdragonConf->new(segs=>['0,0,0','2,0,0','3,1,2']);
#   $conf->expand(0, 1)->print;  # alt
#   $conf->expand(1, 1)->print;  # alt
#   $conf->expand(2, 1)->print;  # alt
# }




#------------------------------------------------------------------------------
# alt_hullstouch

# MyTerdragonConf->initial->alt_expand9(0)->print;
# MyTerdragonConf->initial->rotate(1)->alt_expand9(0)->print;
# MyTerdragonConf->initial->alt_expand9(2)->print;

# MyTerdragonConf->new
#   (name => 'R triangle',
#    segs => [ '0,0,0',       # below m
#              '1,1,2',
#              # '1,-1,1',
#            ])->alt_expand9(0)->print;

is ($alt_hullstouch_segs->num_segs, 10+3);
{
  # $alt_hullstouch_segs expand to within self
  # $alt_hullstouch_segs->print;
  foreach my $digit (0..8) {
    my $exp = $alt_hullstouch_segs->alt_expand9($digit)
      ->set_name("alt_hullstouch_segs expand $digit");
    $exp->print;
    ok($alt_hullstouch_segs->is_subset($exp),
       "alt_hullstouch_segs expand is subset, digit=$digit");

    $exp = $exp->intersection($alt_hullstouch_segs);
    ok($exp->equals($alt_hullstouch_segs),
       "alt_hullstouch_segs expand intersection, digit=$digit");
  }
}
{
  my $conf = $alt_hullstouch_segs->clone
    ->set_name("alt_hullstouch_segs self-within");
  $conf->print;
  $conf->{'within'} = $alt_hullstouch_segs;
  foreach my $digit (0..8) {
    my $exp = $conf->alt_expand9($digit);
    ok($exp->equals($conf), "$conf->{'name'}, alt digit=$digit");
  }
}

&fAltBpred_hullstouch_init(verbose => 1, verbose_print => \&diag);
diag "alt_hullstouch state_table length ",
  scalar(@alt_hullstouch_state_to_conf);


is ($alt_hullstouch_state_table[0]->[0], 1);
is ($alt_hullstouch_state_table[1]->[8], 15);
is ($alt_hullstouch_state_table[15]->[0], 17);
is ($alt_hullstouch_state_table[17]->[0], 17);
ok (  $fAltRpred_hullstouch_flat->is_accepting(0));
ok (  $fAltRpred_hullstouch_flat->is_accepting(1));
ok (  $fAltRpred_hullstouch_flat->is_accepting(15));
ok (! $fAltRpred_hullstouch_flat->is_accepting(17));
ok (  $fAltNonRpred_hullstouch_flat->is_accepting(17));
$alt_hullstouch_state_to_conf[0]->print;
$alt_hullstouch_state_to_conf[15]->print;
$alt_hullstouch_state_to_conf[17]->print;

{
  my $f = FLAT::Regex->new('0 8 0 0 0*')->as_dfa;
  ok ($f->is_subset_of($fAltNonRpred_hullstouch_flat));
}
{
  my $f = $fAltNonRpred_hullstouch_flat->MyFLAT::minimize(verbose=>1);
  is($f->num_states, 11);
}

{
  my $conf = $alt_hullstouch_segs->clone;
  $conf->{'within'} = $alt_hullstouch_segs;
  ok (! MyTerdragon::hullstouch_is_boundary($conf,0));
  ok (! MyTerdragon::hullstouch_is_boundary($conf,1));
}

is($fAltRpred_hullstouch_flat->num_states, 43);  # as stated in dragon.tex
is($fAltLpred_hullstouch_flat->num_states, 43);
is($fAltBpred_hullstouch_flat->num_states, 43);
is($fAltRpred_hullstouch_flat->MyFLAT::num_accepting, 28);
is($fAltLpred_hullstouch_flat->MyFLAT::num_accepting, 28);
is($fAltBpred_hullstouch_flat->MyFLAT::num_accepting, 42);

# fNonB are all eventually accepting already
ok ($fAltNonBpred_hullstouch_flat->equals(MyFLAT::eventually_accepting($fAltNonBpred_hullstouch_flat)));

ok($fAltNonRpred_hullstouch_flat->is_subset_of(AltNonRpred9_flat_even()));
ok(AltRpred9_flat_even()->is_subset_of($fAltRpred_hullstouch_flat));

# ok($fAltNonRpred_hullstouch_flat->equals( fAltNonRpred_flat() ));
# ok($fAltNonLpred_hullstouch_flat->equals( fAltNonLpred_flat() ));
# ok($fAltNonBpred_hullstouch_flat->equals( fAltNonBpred_flat() ));
# 
# ok($fAltRpred_hullstouch_flat->equals( fAltRpred_flat() ));
# ok($fAltLpred_hullstouch_flat->equals( fAltLpred_flat() ));
# ok($fAltBpred_hullstouch_flat->equals( fAltBpred_flat() ));
# 
# ok($fAltRpred_hullstouch_flat
#    ->equals(MyFLAT::rename_accepting_last($fAltRpred_hullstouch_flat)));
# ok($fAltLpred_hullstouch_flat
#    ->equals(MyFLAT::rename_accepting_last($fAltLpred_hullstouch_flat)));
# ok($fAltBpred_hullstouch_flat
#    ->equals(MyFLAT::rename_accepting_last($fAltBpred_hullstouch_flat)));


#------------------------------------------------------------------------------
# $RpointPred_segs()

# $RpointPred_segs->print;
foreach my $digit (0..2) {
  my $exp = $RpointPred_segs->expand($digit);
  # $exp->print;
  ok ($RpointPred_segs->is_subset($exp),
      "RpointPred_segs expand digit=$digit within itself");
}

RpointPred_init();
is (scalar(@RpointPred_state_to_conf), 39,
    "RpointPred_state_to_conf num states");

{
  # segments common to all accepting states
  my $intersection;
  foreach my $state (0 .. $#RpointPred_state_to_conf) {
    ### $state
    my $conf = $RpointPred_state_to_conf[$state];
    next if MyTerdragon::RpointPred_is_boundary($conf);
    if (defined $intersection) {
      $intersection = $intersection->intersection($conf);
    } else {
      $intersection = $conf;
    }
    $conf->print;
  }
  $intersection->print;
}

#------------------------------------------------------------------------------
# RpointPred()
# RpointPred_flat()

ok (RpointPred(3));
ok (RpointPred(5));

{
  # RpointPred() vs RpointPred_flat()
  RpointPred_init();
  my $bad = 0;
 K: foreach my $k (0 .. 5) {
    foreach my $n (0 .. 3**$k) {
      my $rpoint_by_func = RpointPred($n) ? 1 : 0;
      my $str = cnv($n,10,3);
      $str = sprintf '%0*s', $k, $str;
      { my $rpoint_by_flat = $RpointPred_flat->contains($str) ? 1 : 0;
        unless ($rpoint_by_func == $rpoint_by_flat) {
          diag "oops RpointPred k=$k N=$n str=$str func $rpoint_by_func flat $rpoint_by_flat\n";
          last K if $bad++ > 10;
        }
      }
      { my $rpoint_by_regexp = ($str =~ /^$MyTerdragon::RpointPred_regexp$/o ? 1 : 0);
        unless ($rpoint_by_func == $rpoint_by_regexp) {
          diag "oops RpointPred k=$k N=$n str=$str func $rpoint_by_func regexp $rpoint_by_regexp\n";
          last K if $bad++ > 10;
        }
      }
    }
  }
  is ($bad, 0);
}


#------------------------------------------------------------------------------
# Rpred_flat(), Lpred_flat()

{
  # Rpred any 11, 12, 20 means not right boundary
  require FLAT::Regex;
  my $f = FLAT::Regex->new('(0|1|2)* (11 | 12 | 20) (0|1|2)*')->as_dfa;
  ok($f->equals(NonRpred_flat()), "NonRpred by regex");
}
{
  # Lpred any 02, 10, 11 means not left boundary
  require FLAT::Regex;
  my $f = FLAT::Regex->new('(0|1|2)* (02 | 10 | 11) (0|1|2)*')->as_dfa;
  ok($f->equals(NonLpred_k_flat()), "NonLpred_k by regex");
}

{
  # Rpred(), Lpred() vs flats
  my $bad = 0;
 K: foreach my $k (0 .. 5) {
    foreach my $n (0 .. 3**$k) {
      my $str = cnv($n,10,3);
      $str = sprintf '%0*s', $k, $str;
      {
        my $r_by_func = Rpred_of_ternary_str($str) ? 1 : 0;
        { my $r_by_flat = Rpred_flat()->contains($str) ? 1 : 0;
          unless ($r_by_func == $r_by_flat) {
            diag "oops Rpred k=$k N=$n str=$str func $r_by_func flat $r_by_flat\n";
            last K if $bad++ > 10;
          }
        }
        { my $r_by_regexp = ($str =~ /^$MyTerdragon::Rpred_regexp$/o ? 1 : 0);
          unless ($r_by_func == $r_by_regexp) {
            diag "oops Rpred k=$k N=$n str=$str func $r_by_func regexp $r_by_regexp\n";
            last K if $bad++ > 10;
          }
        }
        { my $str_LtoH = reverse $str;
          my $r_by_LtoH_regexp = ($str_LtoH =~ /^$MyTerdragon::Rpred_LtoH_regexp$/o ? 1 : 0);
          unless ($r_by_func == $r_by_LtoH_regexp) {
            diag "oops Rpred k=$k N=$n str=$str LtoH $str_LtoH func $r_by_func regexp $r_by_LtoH_regexp\n";
            last K if $bad++ > 10;
          }
        }
        { my $non_r_by_regexp = ($str =~ /$MyTerdragon::NonRpred_regexp/o ? 1 : 0);
          unless ($r_by_func == 1-$non_r_by_regexp) {
            diag "oops NonRpred k=$k N=$n str=$str func $r_by_func regexp $non_r_by_regexp\n";
            last K if $bad++ > 10;
          }
        }
      }

      {
        my $l_by_func = Lpred_k_of_ternary_str($str) ? 1 : 0;
        my $l_by_flat = Lpred_k_flat()->contains($str) ? 1 : 0;
        unless ($l_by_func == $l_by_flat) {
          diag "oops Lpred k=$k N=$n str=$str func $l_by_func flat $l_by_flat\n";
          last K if $bad++ > 10;
        }
      }
    }
  }
  is($bad,0);
}


#------------------------------------------------------------------------------
# hullstouch

is ($hullstouch_segs->num_segs, 10+1);
{
  # $hullstouch_segs expand to within self
  my $conf = $hullstouch_segs;
  foreach my $digit (0,1,2) {
    ok($conf->is_subset( $conf->expand($digit) ),
       "$conf->{'name'}, digit=$digit");
  }
}

&fBpred_hullstouch_init(verbose => 1, verbose_print => \&diag);
diag("hullstouch state_table length ",scalar(@hullstouch_state_to_conf));

is($fRpred_hullstouch_flat->num_states, 31);  # as stated in dragon.tex
is($fLpred_hullstouch_flat->num_states, 31);
is($fBpred_hullstouch_flat->num_states, 31);

ok($fNonRpred_hullstouch_flat->equals( fNonRpred_flat() ));
ok($fNonLpred_hullstouch_flat->equals( fNonLpred_flat() ));
ok($fNonBpred_hullstouch_flat->equals( fNonBpred_flat() ));

ok($fRpred_hullstouch_flat->equals( fRpred_flat() ));
ok($fLpred_hullstouch_flat->equals( fLpred_flat() ));
ok($fBpred_hullstouch_flat->equals( fBpred_flat() ));

ok($fRpred_hullstouch_flat
   ->equals(MyFLAT::rename_accepting_last($fRpred_hullstouch_flat)));
ok($fLpred_hullstouch_flat
   ->equals(MyFLAT::rename_accepting_last($fLpred_hullstouch_flat)));
ok($fBpred_hullstouch_flat
   ->equals(MyFLAT::rename_accepting_last($fBpred_hullstouch_flat)));


#------------------------------------------------------------------------------
# hullsaround

{
  # expand to within self
  my $conf = $hullsaround_segs_all;
  foreach my $digit (0,1,2) {
    ok($conf->is_subset( $conf->expand($digit) ),
       "$conf->{'name'}, digit=$digit");
  }
}
{
  my $c = $hullsaround_segs_inner;
  foreach my $digit (0,1,2) {
    # $c->expand($digit)->print;
    ok ($c->is_subset( $c->expand($digit) ),
        "$c->{name} digit=$digit");
  }
}

is($hullsaround_segs_inner->intersection($hullsaround_segs_outer)->num_segs, 0);
# $hullsaround_segs_outer->print;
# $hullsaround_segs_inner->print;
# $hullsaround_segs_all->print;


&fBpred_hullsaround_init(verbose => 1);
diag("hullsaround state_table length ",scalar(@hullsaround_state_to_conf));

ok($fNonRpred_hullsaround_flat->equals( fNonRpred_flat() ));
ok($fNonLpred_hullsaround_flat->equals( fNonLpred_flat() ));
ok($fNonBpred_hullsaround_flat->equals( fNonBpred_flat() ));

ok($fRpred_hullsaround_flat->equals( fRpred_flat() ));
ok($fLpred_hullsaround_flat->equals( fLpred_flat() ));
ok($fBpred_hullsaround_flat->equals( fBpred_flat() ));

ok($fRpred_hullsaround_flat
   ->equals(MyFLAT::rename_accepting_last($fRpred_hullsaround_flat)));
ok($fLpred_hullsaround_flat
   ->equals(MyFLAT::rename_accepting_last($fLpred_hullsaround_flat)));
ok($fBpred_hullsaround_flat
   ->equals(MyFLAT::rename_accepting_last($fBpred_hullsaround_flat)));

ok($fRpred_hullsaround_flat->equals($fRpred_hullstouch_flat));
ok($fLpred_hullsaround_flat->equals($fLpred_hullstouch_flat));
ok($fBpred_hullsaround_flat->equals($fBpred_hullstouch_flat));



#------------------------------------------------------------------------------
# AltRpred vs path

{
  my $f = AltRpred3_flat();
}


#------------------------------------------------------------------------------
# EpredRpair, EpredLpair

{
  my $f = EpredR_flat()->MyFLAT::concat(FLAT::Regex->new('022*'));
  ok (EpredRpair_flat()->equals($f));
}
{
  my $f = EpredL_flat()->MyFLAT::concat(FLAT::Regex->new('2*1'));
  ok (EpredLpair_flat()->equals($f));
}


#------------------------------------------------------------------------------
# Epred runs

{
  # EpredTwelve vs regexp
  require FLAT::Regex;
  #                                      6                 5 4 3 2 1
  my $f = FLAT::Regex->new('(0|1|2)* 1 0* (1 2* 2 0*)* 1 2* 1 2* 2 1 1')
    #                                ^     ^----^      ^    ^   |-----|
    ->as_dfa
    ->MyFLAT::set_name('EpredTwelve by regex');

  ok ($f->equals(EpredTwelve_LtoH_flat()->MyFLAT::reverse),
      "EpredTwelve by regex");

  # $f->MyFLAT::reverse->MyFLAT::minimize->MyFLAT::view (flow => 'left');
  # EpredTwelve_LtoH_flat()->MyFLAT::view (flow => 'left');
  # MyFLAT::FLAT_check_is_equal($f,EpredTwelve_LtoH_flat()->MyFLAT::reverse);
}
{
  # EpredTwelve by n to n+11
  require MyFLAT;
  my $f = Epred_flat();
  for (1 .. 11) {
    $f = Epred_flat()->intersect($f->MyFLAT::digits_increment(add => -1))
      ->MyFLAT::minimize;
  }
  ok($f->equals(EpredTwelve_LtoH_flat()->MyFLAT::reverse));
}

{
  my $f = Epred_flat();
  for (1 .. 11) {
    $f = Epred_flat()->intersect($f->MyFLAT::digits_increment)
      ->MyFLAT::minimize(verbose=>1);
  }
  # MyFLAT::FLAT_show_breadth($f, 8);
  my $max_k = 15;
  is(join(',', FLAT_count_contains($f,$max_k)),
     '0,0,0,0,0,0,1,6,25,90,301,966,3025,9330,28501,86526');

  # 12 increments, no run >=13 consecutive enclosures
  $f = Epred_flat()->intersect($f->MyFLAT::digits_increment);
  ok($f->is_empty);
}


#------------------------------------------------------------------------------
# fVisits_flat

{
  # Visits_k_flat()

  my $one  = MyTerdragon::Visits_k_flat(1);
  my $two  = MyTerdragon::Visits_k_flat(2);
  my $three = MyTerdragon::Visits_k_flat(3)->MyFLAT::minimize(verbose=>1);;

  ok($one->intersect($two)->is_empty, 'Visits 1,2 disjoint');
  ok($one->intersect($three)->is_empty, 'Visits 1,3 disjoint');
  ok($two->intersect($three)->is_empty, 'Visits 2,3 disjoint');

  # $three->MyFLAT::view;

  # Visits=3 has accepting sink once away from boundary.
  # Other visits can always become enclosed in a sub-curve.
  # 
  is_deeply([$one->MyFLAT::get_accepting_sinks],[]);
  is_deeply([$two->MyFLAT::get_accepting_sinks],[]);
  is ($one->MyFLAT::num_accepting_sinks, 0);
  is ($two->MyFLAT::num_accepting_sinks, 0);
  isnt ($three->MyFLAT::num_accepting_sinks, 0);
}

{
  # f not 3^k any 11 means non-boundary
  foreach my $regex ('(0|1|2)* 11 (0|1|2)*',
                     '(0|1|2)* 02 (0|1|2)* 12 (0|1|2)*',
                     '(0|1|2)* 02 (0|1|2)* 20 (0|1|2)*',
                     '(0|1|2)* 020 (0|1|2)*',
                     '(0|1|2)* 10 (0|1|2)* 12 (0|1|2)*',
                     '(0|1|2)* 10 (0|1|2)* 20 (0|1|2)*',
                    ) {
    my $f = FLAT::Regex->new($regex)->as_dfa;
    ok(Bpred_k_flat()->intersect($f)->is_empty);
  }
}

{
  # fVisits_exact_flat()

  my $one  = MyTerdragon::fVisits_exact_flat(1);
  my $two  = MyTerdragon::fVisits_exact_flat(2);
  my $three = MyTerdragon::fVisits_exact_flat(3);

  #
  is_deeply([$one->MyFLAT::get_accepting_sinks],[]);
  is_deeply([$two->MyFLAT::get_accepting_sinks],[]);
  is ($one->MyFLAT::num_accepting_sinks, 0);
  is ($two->MyFLAT::num_accepting_sinks, 0);
  isnt ($three->MyFLAT::num_accepting_sinks, 0);

  #  $three->MyFLAT::view;

  # MyFLAT::FLAT_show_breadth($one->intersect($two),4);
  # ok($one->intersect($two)->is_empty, 'fVisits exact 1,2 disjoint');
  # ok($one->intersect($three)->is_empty, 'fVisits exact 1,3 disjoint');
  # ok($two->intersect($three)->is_empty, 'fVisits exact 2,3 disjoint');
}
{
  # MyTerdragon::fVisits_nonexact_2_flat()->MyFLAT::view;

  my $one  = MyTerdragon::fVisits_flat(1);
  my $two  = MyTerdragon::fVisits_flat(2);
  my $three = MyTerdragon::fVisits_flat(3);
  # $one->MyFLAT::view;
  # $two->MyFLAT::view;
  # $three->MyFLAT::view;

  # ok($one->intersect($two)->is_empty);
  # ok($one->intersect($three)->is_empty);
  # ok($two->intersect($three)->is_empty);
}


#------------------------------------------------------------------------------
# Convex Hull

{
  my ($x,$y) = P1(6);
  is ($x, 2);
  is ($y, 0);
}
{
  my ($x,$y) = P7(6);
  is ($x, -33);
  is ($y, 9);
}

# p(k+4) = w3*p(k)
foreach my $i (0 .. 11) {
  my ($x,$y) = p($i);
  ($x+$y) % 2 == 0 or die;
  ($x,$y) = xy_rotate_plus120($x,$y);
  my ($wx,$wy) = p($i+4);
  is ($x, $wx);
  is ($y, $wy); # or die "$x,$y want $wx,$wy";
}


#------------------------------------------------------------------------------
# Area Graph

terdragon_Graph(0);

# $graph is a Graph.pm.
# Return a new graph where each vertex is replaced by a triangle.
# Existing edges become a vertex in common between the new triangles.
#
sub Graph_vertices_to_triangles {
  my ($graph) = @_;
  my $new = Graph->new(undirected => $graph->is_undirected);
  if (defined (my $name = $graph->get_graph_attribute('name'))) {
    $graph->set_graph_attribute (name => "$name - Triangles");
  }
  my %v_to_t;
  my $upto = 1;
  foreach my $edge ($graph->edges) {
    my ($v1,$v2) = @$edge;
    my $t = $upto++;
    $new->add_vertex($t);
    push @{$v_to_t{$v1}}, $t;  # vertex in common for each edge
    push @{$v_to_t{$v2}}, $t;
  }
  foreach my $v ($graph->vertices) {
    $v_to_t{$v} ||= [];
    while (scalar(@{$v_to_t{$v}}) < 3) {
      my $t = $upto++;
      $new->add_vertex($t);
      push @{$v_to_t{$v}}, $t;   # triangle for each vertex
    }
  }
  foreach my $aref (values %v_to_t) {
    $new->add_cycle(@$aref);
  }
  return $new;
}

# area graph is the turn tree with vertices replaced by triangles
require MyGraphs;
foreach my $k (0 .. 5) {
  my $turn = triangle_turn_tree_Graph($k);
  my $area = triangle_area_Graph($k);
  my $triangles = Graph_vertices_to_triangles($turn);
  my $ok = MyGraphs::Graph_is_isomorphic($triangles, $area) || $k==0;
  ok ($ok, "k=$k");
  if (! $ok) {
    # MyGraphs::Graph_view($area);
    # MyGraphs::Graph_view($turn);
    # MyGraphs::Graph_view($triangles);
  }
}

# area graph is the turn tree padded to degree-3 and take line graph
require MyGraphs;
foreach my $k (0 .. 5) {
  my $area = triangle_area_Graph($k);
  my $t = triangle_turn_tree_Graph($k);
  MyGraphs::Graph_pad_degree($t, 3);
  $t = MyGraphs::Graph_line_graph($t);
  my $ok = MyGraphs::Graph_is_isomorphic($t, $area) || $k==0;
  ok ($ok, "k=$k line graph of padded");
  if (! $ok) {
    # MyGraphs::Graph_view($area);
    # MyGraphs::Graph_view($t);
  }
}

#------------------------------------------------------------------------------
# sequences

{
  # R
  my @want = (1, 3, 6, 12, 24, 48);
  my @got = map {R($_)} 0 .. $#want;
  is_deeply(\@got,\@want, 'R');
}

{
  # D
  my @want = (0, 0, 2, 6, 14, 30, 62, 126, 254);
  my @got = map {D($_)} 0 .. $#want;
  is_deeply(\@got,\@want, 'D');
}

{
  # P         0  1  2  3  4
  my @want = (2, 4, 8, 18, 44, 114, 308, 858);
  my @got = map {P($_)} 0 .. $#want;
  is_deeply(\@got,\@want, 'P');
}

#------------------------------------------------------------------------------
# EpredR_flat(), EpredL_flat()

{
  my $EpredR_LtoH_state_aref = [ [1,2,0],   # rm0
                                 [1,0,3],   # rm1
                                 [2,2,2],   # non
                                 [3,3,3],   # enc
                               ];
  require MyFLAT;
  my $EpredR_LtoH_flat = MyFLAT::aref_to_FLAT_DFA($EpredR_LtoH_state_aref,
                                                  name => 'EpredR LtoH');
  ok ($EpredR_LtoH_flat->MyFLAT::reverse->equals(EpredR_flat()), "EpredR");
}
{
  my $EpredL_LtoH_state_aref = [ [2,1,0],   # lm0
                                 [1,3,0],   # lm1
                                 [2,2,2],   # non
                                 [3,3,3],   # enc
                               ];
  require MyFLAT;
  my $EpredL_LtoH_flat = MyFLAT::aref_to_FLAT_DFA($EpredL_LtoH_state_aref,
                                                  name => 'EpredL k LtoH');
  ok ($EpredL_LtoH_flat->MyFLAT::reverse->equals(EpredL_flat()), "EpredL");
}

# Subsets of the fully enclosed R,L.
# Epred is the last segment number, whereas NonR,NonL are all segments.
ok(EpredR_flat()->is_subset_of(NonRpred_flat() ));
ok(EpredL_flat()->is_subset_of(NonLpred_k_flat() ));


#------------------------------------------------------------------------------
# turn()

is(turn(1), 1);
is(turn(2), -1);
is(turn(3), 1);

{
  my $path = Math::PlanePath::TerdragonCurve->new;
  my $seq = Math::NumSeq::PlanePathTurn->new(planepath_object=>$path,
                                             turn_type => 'LSR');
  my $bad = 0;
  foreach (1 .. 500) {
    my ($n,$by_seq) = $seq->next;
    my $by_func = turn($n);
    unless ($by_seq == $by_func) {
      diag "n=$n turn()=$by_func seq=$by_seq";
      last if $bad++ > 10;
    }
  }
  is($bad, 0);
}

#------------------------------------------------------------------------------
# R2sides, L2sides

ok(R1sides_flat()  ->is_subset_of(Rpred_flat() ));
ok(R2sides_flat()  ->is_subset_of(Rpred_flat() ));

ok(R1sides_flat()->intersect(R2sides_flat())->is_empty,  'disjoint');
ok(R1sides_flat()->intersect(NonRpred_flat())->is_empty,'disjoint');
ok(R2sides_flat()->intersect(NonRpred_flat())->is_empty,'disjoint');

{
  my $f = R1sides_flat()->union(R2sides_flat(),NonRpred_flat())->as_dfa;
  ok($f->equals($all_ternary), 'R1sides, R2sides, NonR is everything');
}

ok(L1sides_k_flat()->is_subset_of(Lpred_k_flat() ));
ok(L2sides_k_flat()->is_subset_of(Lpred_k_flat() ));

ok(L1sides_k_flat()->intersect(L2sides_k_flat())->is_empty,  'disjoint');
ok(L1sides_k_flat()->intersect(NonLpred_k_flat())->is_empty,'disjoint');
ok(L2sides_k_flat()->intersect(NonLpred_k_flat())->is_empty,'disjoint');

{
  my $f = L1sides_k_flat()->union(L2sides_k_flat(),NonLpred_k_flat())->as_dfa;
  ok($f->equals($all_ternary), 'L1sides, L2sides, NonL is everything');
}



#------------------------------------------------------------------------------
# Spred, Dpred, Tpred

ok(Spred_k_flat()->intersect(Dpred_k_flat())->is_empty, 'disjoint');
ok(Spred_k_flat()->intersect(Tpred_k_flat())->is_empty, 'disjoint');
ok(Dpred_k_flat()->intersect(Tpred_k_flat())->is_empty, 'disjoint');

ok(SpredRight_k_flat()->intersect(DpredRight_k_flat())->is_empty, 'disjoint');
ok(SpredRight_k_flat()->intersect(TpredRight_k_flat())->is_empty, 'disjoint');
ok(DpredRight_k_flat()->intersect(TpredRight_k_flat())->is_empty, 'disjoint');

ok(SpredLeft_flat()->intersect(DpredLeft_flat())->is_empty, 'disjoint');
ok(SpredLeft_flat()->intersect(TpredLeft_flat())->is_empty, 'disjoint');
ok(DpredLeft_flat()->intersect(TpredLeft_flat())->is_empty, 'disjoint');

ok (Spred_k_flat()->union(Dpred_k_flat(),Tpred_k_flat())->as_dfa
    ->equals($all_ternary),
    'Spred, Dpred, Tpred is everything');
ok (SpredRight_k_flat()->union(DpredRight_k_flat(),
                               TpredRight_k_flat())->as_dfa
    ->equals(TurnRpred_flat()),
    'SpredRight, DpredRight, TpredRight is TurnRpred');
ok (SpredLeft_flat()->union(DpredLeft_flat(),
                            TpredLeft_flat())->as_dfa
    ->equals(TurnLpred_flat()),
    'SpredLeft, DpredLeft, TpredLeft is TurnLpred');

ok(SpredLeft_flat()->is_subset_of(Spred_k_flat()));
ok(DpredLeft_flat()->is_subset_of(Dpred_k_flat()));
ok(TpredLeft_flat()->is_subset_of(Tpred_k_flat()));

ok(SpredLeft_flat()->is_subset_of(TurnLpred_flat() ));
ok(DpredLeft_flat()->is_subset_of(TurnLpred_flat() ));
ok(TpredLeft_flat()->is_subset_of(TurnLpred_flat() ));

ok(SpredRight_k_flat()->is_subset_of(Spred_k_flat()));
ok(DpredRight_k_flat()->is_subset_of(Dpred_k_flat()));
ok(TpredRight_k_flat()->is_subset_of(Tpred_k_flat()));

ok(SpredRight_k_flat()->is_subset_of(TurnRpred_flat() ));
ok(DpredRight_k_flat()->is_subset_of(TurnRpred_flat() ));
ok(TpredRight_k_flat()->is_subset_of(TurnRpred_flat() ));

ok(DpredLeft_flat()->contains('11'));

{
  my $path = Math::PlanePath::TerdragonCurve->new;
  my $seq = Math::NumSeq::PlanePathTurn->new(planepath_object=>$path,
                                             turn_type => 'LSR');
  my $bad = 0;
 K: foreach my $k (0 .. 5) {
    my ($n_lo,$n_hi) = $path->level_to_n_range($k);
    foreach my $n (0 .. 3**$k) {
      my $str = cnv($n,10,3);
      $str = sprintf '%0*s', $k, $str;

      my $turn = $n>0 && $seq->ith($n);
      my ($x,$y) = $path->n_to_xy($n);
      my @n_list = $path->xy_to_n_list($x,$y);
      @n_list = grep {$_ <= $n_hi} @n_list;  # within this level k

      # SpredLeft
      {
        my $got  = SpredLeft_flat()->contains($str) ? 1 : 0;
        my $want = @n_list==1 && $turn==1;
        unless ($got == $want) {
          diag "oops SpredLeft k=$k N=$n str=$str got $got want $want (turn $turn)\n";
          last K if $bad++ > 10;
        }
      }
      # SpredRight
      {
        my $got  = SpredRight_k_flat()->contains($str) ? 1 : 0;
        my $want = @n_list==1 && $turn==-1;
        unless ($got == $want) {
          diag "oops SpredRight k=$k N=$n str=$str got $got want $want (turn $turn)\n";
          last K if $bad++ > 10;
        }
      }
      # Spred
      {
        my $got  = Spred_k_flat()->contains($str) ? 1 : 0;
        my $want = @n_list==1;
        unless ($got == $want) {
          diag "oops Spred k=$k N=$n str=$str got $got want $want (turn $turn)\n";
          last K if $bad++ > 10;
        }
      }

      # DpredLeft
      {
        my $got  = DpredLeft_flat()->contains($str) ? 1 : 0;
        my $want = @n_list==2 && $turn==1;
        unless ($got == $want) {
          diag "oops DpredLeft k=$k N=$n str=$str got $got want $want (turn $turn)\n";
          last K if $bad++ > 10;
        }
      }
      # DpredRight
      {
        my $got  = DpredRight_k_flat()->contains($str) ? 1 : 0;
        my $want = @n_list==2 && $turn==-1;
        unless ($got == $want) {
          diag "oops DpredRight k=$k N=$n str=$str got $got want $want (turn $turn)\n";
          last K if $bad++ > 10;
        }
      }
      # Dpred
      {
        my $got  = Dpred_k_flat()->contains($str) ? 1 : 0;
        my $want = @n_list==2;
        unless ($got == $want) {
          diag "oops Dpred k=$k N=$n str=$str got $got want $want (turn $turn)\n";
          last K if $bad++ > 10;
        }
      }

      # TpredLeft
      {
        my $got  = TpredLeft_flat()->contains($str) ? 1 : 0;
        my $want = @n_list==3 && $turn==1;
        unless ($got == $want) {
          diag "oops TpredLeft k=$k N=$n str=$str got $got want $want (turn $turn)\n";
          last K if $bad++ > 10;
        }
      }
      # TpredRight
      {
        my $got  = TpredRight_k_flat()->contains($str) ? 1 : 0;
        my $want = @n_list==3 && $turn==-1;
        unless ($got == $want) {
          diag "oops TpredRight k=$k N=$n str=$str got $got want $want (turn $turn)\n";
          last K if $bad++ > 10;
        }
      }
      # Tpred
      {
        my $got  = Tpred_k_flat()->contains($str) ? 1 : 0;
        my $want = @n_list==3;
        unless ($got == $want) {
          diag "oops Tpred k=$k N=$n str=$str got $got want $want (turn $turn)\n";
          last K if $bad++ > 10;
        }
      }
    }
  }
  is($bad,0);
}

#------------------------------------------------------------------------------
# fRpred

ok (fRpred_flat()->equals(fNonRpred_flat()->complement));
ok (fLpred_flat()->equals(fNonLpred_flat()->complement));
ok (fBpred_flat()->equals(fNonBpred_flat()->complement));

# f are all accepting prefixes
ok (fRpred_flat()->equals(MyFLAT::prefix_accepting(fRpred_flat())));
ok (fLpred_flat()->equals(MyFLAT::prefix_accepting(fLpred_flat())));
ok (fBpred_flat()->equals(MyFLAT::prefix_accepting(fBpred_flat())));

# fNon are all eventually accepting already
ok (fNonRpred_flat()->equals(MyFLAT::eventually_accepting(fNonRpred_flat())));
ok (fNonLpred_flat()->equals(MyFLAT::eventually_accepting(fNonLpred_flat())));
ok (fNonBpred_flat()->equals(MyFLAT::eventually_accepting(fNonBpred_flat())));

# fNon are subsets of the finite iterations, accepting the extra Dpreds
ok (fNonRpred_flat()->is_subset_of(NonRpred_flat()));
ok (fNonLpred_flat()->is_subset_of(NonLpred_k_flat()));

{
  # intersection both fRpred and fLpred is only f=0 and f=1

  # eventually non-accepting
  my $got = fRpred_flat()->intersect(fLpred_flat());
  $got = $got->complement;
  $got = $got->MyFLAT::eventually_accepting;
  $got = $got->complement;

  my $want = FLAT::Regex->new('0*|2*')->as_dfa;
  ok ($got->equals($want));
  # MyFLAT::FLAT_check_is_equal($got,$want);
}

# use constant::defer DpredRight_k_borrow_flat => sub {
#   require MyFLAT;
#   my $f = MyFLAT::FLAT_concat(
#                               FLAT::Regex->new('0'));
#   $f->{'name'} = 'DpredRight';
#   return $f;
# };
# use constant::defer DpredLeft_borrow_flat => sub {
#   require MyFLAT;
#   my $f = MyFLAT::FLAT_concat(L2sides_k_flat(),
#                               FLAT::Regex->new('1'));
#   $f->{'name'} = 'DpredLeft';
#   return $f;
# };

# fRpred vs theorem statement
my $fRpred_case_DpredLeft_flat
  = R2sides_flat()
  ->MyFLAT::concat(FLAT::Regex->new('1 0*'))
  ->union(R2sides_flat()->MyFLAT::concat(FLAT::Regex->new('0 2*')))
  ->as_dfa
  ->MyFLAT::prefix
  ->MyFLAT::set_name('fRpred_case_DpredLeft');

my $fRpred_by_cases
  = Rpred_flat()
  ->union($fRpred_case_DpredLeft_flat)
  ->as_dfa;
# $f = $f->MyFLAT::prefix;

ok (fRpred_flat()->equals($fRpred_by_cases), "fRpred by cases");

# fLpred vs theorem statement
my $fLpred_case_DpredRight_flat
  = L2sides_k_flat()
  ->MyFLAT::concat(FLAT::Regex->new('2 0*')->as_dfa)
  ->union(L2sides_k_flat()->MyFLAT::concat(FLAT::Regex->new('1 2*')))
  ->as_dfa
  ->MyFLAT::prefix
  ->MyFLAT::set_name('fLpred_case_DpredRight');
my $fLpred_by_cases
  = Lpred_k_flat()->union($fLpred_case_DpredRight_flat)
  ->as_dfa
  ->MyFLAT::set_name('fLpred_by_cases');
ok (fLpred_flat()->equals($fLpred_by_cases), "fLpred by cases");

{
  my $f = SpredLeft_flat()
    ->MyFLAT::concat(FLAT::Regex->new('0*')->as_dfa)
    ->MyFLAT::prefix;
  ok ($f->is_subset_of(fRpred_flat()), "SpredLeft subset of fRpred");
}
{
  my $f = SpredRight_k_flat()
    ->MyFLAT::concat(FLAT::Regex->new('0*')->as_dfa)
    ->MyFLAT::prefix;
  ok ($f->is_subset_of(fLpred_flat()), "SpredRight subset of fLpred");
}

# all triple-visited points detected as NonRpred,NonLpred after one more 0 digit
{
  my $f = Tpred_k_flat()->MyFLAT::concat(FLAT::Regex->new('0')->as_dfa);
  ok ($f->is_subset_of(NonRpred_flat()), "Tpred subset of NonRpred");
  ok ($f->is_subset_of(NonLpred_k_flat()), "Tpred subset of NonLpred");
}

# all triple-visited points detected as fNonRpred,fNonLpred after one more 0 digit

# 0220 N=24 is fRpred_flat()
# 1020 N=33 is fRpred_flat()
{
  my $f = Tpred_k_flat()->MyFLAT::concat(FLAT::Regex->new('00')->as_dfa);
  ok ($f->is_subset_of(fNonRpred_flat()), "Tpred subset of fNonRpred");
}
{
  my $f = Tpred_k_flat()->MyFLAT::concat(FLAT::Regex->new('0')->as_dfa);
  ok ($f->is_subset_of(fNonLpred_flat()), "Tpred subset of fNonLpred");

  # $f = MyFLAT::eventually_accepting($f, verbose=>1);
  # my $i = $f->intersect(fRpred_flat())->MyFLAT::minimize;
  #  MyFLAT::view($i);
  # FLAT_show_breadth($i,8);
}

#-------------------------------------------------------------------------------
# fRpred, fLpred vs M2, E2 extra states shown in dragon.tex

{
  my @state_table = ([ 1,2,3 ],   # R = 1
                     [ 1,4,6 ],   # M = 2
                     [ 5,2,3 ],   # E = 3
                     [ 4,6,6 ],   # M2 = 4
                     [ 6,6,5 ],   # E2 = 5
                     [ 6,6,6 ]);  # non
  @state_table = map {[map{$_-1}@$_]} @state_table;  # to 0-based
  my $f = MyFLAT::aref_to_FLAT_DFA(\@state_table);
  ok ($f->equals(fNonRpred_flat()),
      "fNonRpred state table in .tex");
}

my %ternary_digit_reverse = (0=>2,1=>1,2=>0);

# left and right same with digits reversed 0 <-> 2
ok(fLpred_flat()
   ->equals( FLAT_rename(fRpred_flat(),
                         symbols_map => \%ternary_digit_reverse)));
ok(fNonLpred_flat()
   ->equals( FLAT_rename(fNonRpred_flat(),
                         symbols_map => \%ternary_digit_reverse)));


#------------------------------------------------------------------------------
exit 0;
